import torch
from torch.nn import Module
from functools import partial
import warnings

from .kernel_samples import kernel_tensorized, kernel_online, kernel_multiscale

from .sinkhorn_samples import sinkhorn_tensorized
from .sinkhorn_samples import sinkhorn_online
from .sinkhorn_samples import sinkhorn_multiscale

from .kernel_samples import kernel_tensorized as hausdorff_tensorized
from .kernel_samples import kernel_online     as hausdorff_online
from .kernel_samples import kernel_multiscale as hausdorff_multiscale


routines = {
    "sinkhorn" : {
        "tensorized" : sinkhorn_tensorized,
        "online"     : sinkhorn_online,
        "multiscale" : sinkhorn_multiscale,
        },
    "hausdorff" : {
        "tensorized" : hausdorff_tensorized,
        "online"     : hausdorff_online,
        "multiscale" : hausdorff_multiscale,
        },
    "energy" : {
        "tensorized" : partial( kernel_tensorized, name="energy" ),
        "online"     : partial( kernel_online,     name="energy" ),
        "multiscale" : partial( kernel_multiscale, name="energy" ),
    },
    "gaussian" : {
        "tensorized" : partial( kernel_tensorized, name="gaussian" ),
        "online"     : partial( kernel_online,     name="gaussian" ),
        "multiscale" : partial( kernel_multiscale, name="gaussian" ),
    },
    "laplacian" : {
        "tensorized" : partial( kernel_tensorized, name="laplacian" ),
        "online"     : partial( kernel_online,     name="laplacian" ),
        "multiscale" : partial( kernel_multiscale, name="laplacian" ),
    },
}


class SamplesLoss(Module):
    """Creates a criterion that computes distances between sampled measures on a vector space.

    Warning:
        If **loss** is ``"sinkhorn"`` and **reach** is **None** (balanced Optimal Transport),
        the resulting routine will expect measures whose total masses are equal with each other.

    Parameters:
        loss (string, default = ``"sinkhorn"``): The loss function to compute.
            The supported values are:

              - ``"sinkhorn"``: (Un-biased) Sinkhorn divergence, which interpolates
                between Wasserstein (blur=0) and kernel (blur= :math:`+\infty` ) distances.
              - ``"hausdorff"``: Weighted Hausdorff distance, which interpolates
                between the ICP loss (blur=0) and a kernel distance (blur= :math:`+\infty` ).
              - ``"energy"``: Energy Distance MMD, computed using the kernel
                :math:`k(x,y) = -\|x-y\|_2`.
              - ``"gaussian"``: Gaussian MMD, computed using the kernel
                :math:`k(x,y) = \exp \\big( -\|x-y\|_2^2 \,/\, 2\sigma^2)`
                of standard deviation :math:`\sigma` = **blur**.
              - ``"laplacian"``: Laplacian MMD, computed using the kernel
                :math:`k(x,y) = \exp \\big( -\|x-y\|_2 \,/\, \sigma)`
                of standard deviation :math:`\sigma` = **blur**.
        
        p (int, default=2): If **loss** is ``"sinkhorn"`` or ``"hausdorff"``,
            specifies the ground cost function between points.
            The supported values are:

              - **p** = 1: :math:`~~C(x,y) ~=~ \|x-y\|_2`.
              - **p** = 2: :math:`~~C(x,y) ~=~ \\tfrac{1}{2}\|x-y\|_2^2`.
        
        blur (float, default=.05): The finest level of detail that
            should be handled by the loss function - in
            order to prevent overfitting on the samples' locations.
            
            - If **loss** is ``"gaussian"`` or ``"laplacian"``,
              it is the standard deviation :math:`\sigma` of the convolution kernel.
            - If **loss** is ``"sinkhorn"`` or ``"haudorff"``,
              it is the typical scale :math:`\sigma` associated
              to the temperature :math:`\\varepsilon = \sigma^p`.
              The default value of .05 is sensible for input
              measures that lie in the unit square/cube.

            Note that the *Energy Distance* is scale-equivariant, and won't 
            be affected by this parameter.

        reach (float, default=None= :math:`+\infty` ): If **loss** is ``"sinkhorn"``
            or ``"hausdorff"``,
            specifies the typical scale :math:`\\tau` associated
            to the constraint strength :math:`\\rho = \\tau^p`.

        diameter (float, default=None): A rough indication of the maximum
            distance between points, which is used to tune the :math:`\\varepsilon`-scaling
            descent and provide a default heuristic for clustering **multiscale** schemes.
            If **None**, a conservative estimate will be computed on-the-fly.

        scaling (float, default=.5): If **loss** is ``"sinkhorn"``,
            specifies the ratio between successive values
            of :math:`\sigma=\\varepsilon^{1/p}` in the
            :math:`\\varepsilon`-scaling descent.
            This parameter allows you to specify the trade-off between
            speed (**scaling** < .4) and accuracy (**scaling** > .9).

        truncate (float, default=None= :math:`+\infty`): If **backend**
            is ``"multiscale"``, specifies the effective support of
            a Gaussian/Laplacian kernel as a multiple of its standard deviation.
            If **truncate** is not **None**, kernel truncation
            steps will assume that 
            :math:`\\exp(-x/\sigma)` or
            :math:`\\exp(-x^2/2\sigma^2) are zero when 
            :math:`\|x\| \,>\, \\text{truncate}\cdot \sigma`.
            

        cost (function or string, default=None): if **loss** is ``"sinkhorn"``
            or ``"hausdorff"``, specifies the cost function that should
            be used instead of :math:`\\tfrac{1}{p}\|x-y\|^p`:
            
            - If **backend** is ``"tensorized"``, **cost** should be a 
              python function that takes as input a
              (B,N,D) torch Tensor **x**, a (B,M,D) torch Tensor **y**
              and returns a batched Cost matrix as a (B,N,M) Tensor.
            - Otherwise, if **backend** is ``"online"`` or ``"multiscale"``,
              **cost** should be a `KeOps formula <http://www.kernel-operations.io/api/math-operations.html>`_,
              given as a string, with variables ``X`` and ``Y``.
              The default values are ``"Norm2(X-Y)"`` (for **p** = 1) and
              ``"(SqDist(X,Y) / IntCst(2))"`` (for **p** = 2).

        cluster_scale (float, default=None): If **backend** is ``"multiscale"``,
            specifies the coarse scale at which cluster centroids will be computed.
            If **None**, a conservative estimate will be computed from
            **diameter** and the ambient space's dimension, 
            making sure that memory overflows won't take place.

        debias (bool, default=True): If **loss** is ``"sinkhorn"``,
            specifies if we should compute the **unbiased**
            Sinkhorn divergence instead of the classic,
            entropy-regularized "SoftAssign" loss.

        potentials (bool, default=False): When this parameter is set to True,
            the :mod:`SamplesLoss` layer returns a pair of optimal dual potentials
            :math:`F` and :math:`G`, sampled on the input measures,
            instead of differentiable scalar value.
            These dual vectors :math:`(F(x_i))` and :math:`(G(y_j))`
            are encoded as Torch tensors, with the same shape
            as the input weights :math:`(\\alpha_i)` and :math:`(\\beta_j)`.

        verbose (bool, default=False): If **backend** is ``"multiscale"``,
            specifies whether information on the clustering and
            :math:`\\varepsilon`-scaling descent should be displayed
            in the standard output.

        backend (string, default = ``"auto"``): The implementation that
            will be used in the background; this choice has a major impact
            on performance. The supported values are:

              - ``"auto"``: Choose automatically, using a simple
                heuristic based on the inputs' shapes.
              - ``"tensorized"``: Relies on a full cost/kernel matrix, computed
                once and for all and stored on the device memory. 
                This method is fast, but has a quadratic
                memory footprint and does not scale beyond ~5,000 samples per measure.
              - ``"online"``: Computes cost/kernel values on-the-fly, leveraging
                online map-reduce CUDA routines provided by 
                the `pykeops <https://www.kernel-operations.io>`_ library.
              - ``"multiscale"``: Fast implementation that scales to millions
                of samples in dimension 1-2-3, relying on the block-sparse
                reductions provided by the `pykeops <https://www.kernel-operations.io>`_ library.

    """
    def __init__(self, loss="sinkhorn", p=2, blur=.05, reach=None, diameter=None, scaling=.5, truncate=5, cost=None, kernel=None, cluster_scale=None, debias=True, potentials=False, verbose=False, backend="auto"):

        super(SamplesLoss, self).__init__()
        self.loss = loss
        self.backend = backend
        self.p = p
        self.blur = blur
        self.reach = reach
        self.truncate = truncate
        self.diameter = diameter
        self.scaling = scaling
        self.cost = cost
        self.kernel = kernel
        self.cluster_scale = cluster_scale
        self.debias = debias
        self.potentials = potentials
        self.verbose = verbose


    def forward(self, *args):
        """Computes the loss between sampled measures.
        
        Documentation and examples: Soon!
        Until then, please check the tutorials :-)"""
        
        l_x, α, x, l_y, β, y = self.process_args(*args)
        B, N, M, D = self.check_shapes(l_x, α, x, l_y, β, y)

        backend = self.backend  # Choose the backend -----------------------------------------
        if l_x is not None or l_y is not None:
            if backend in ["auto", "multiscale"]: backend = "multiscale"
            else:
                raise ValueError('Explicit cluster labels are only supported with the "auto" and "multiscale" backends.')

        elif backend == "auto":
            if M*N <= 5000**2 : backend = "tensorized"  # Fast backend, with a quadratic memory footprint
            else:
                if D <= 3 and self.loss == "sinkhorn" and M*N > 10000**2 and self.p==2:
                    backend = "multiscale"  # Super scalable algorithm in low dimension     
                else:
                    backend = "online"      # Play it safe, without kernel truncation

        # Check compatibility between the batchsize and the backend --------------------------
        if backend in ["online", "multiscale"]:  # KeOps routines work on single measures
            if B == 1:
                α, x, β, y = α.squeeze(0), x.squeeze(0), β.squeeze(0), y.squeeze(0)
            elif B > 1:
                warnings.warn("'online' and 'multiscale' backends do not support batchsize > 1. " \
                             +"Using 'tensorized' instead: beware of memory overflows!")
                backend = "tensorized"

        if B == 0 and backend == "tensorized":  # tensorized routines work on batched tensors
            α, x, β, y = α.unsqueeze(0), x.unsqueeze(0), β.unsqueeze(0), y.unsqueeze(0)


        # Run --------------------------------------------------------------------------------
        values = routines[self.loss][backend]( α, x, β, y, 
                    p = self.p, blur = self.blur, reach = self.reach, 
                    diameter=self.diameter, scaling = self.scaling, truncate = self.truncate, 
                    cost = self.cost, kernel = self.kernel, cluster_scale=self.cluster_scale,
                    debias = self.debias, potentials = self.potentials,
                    labels_x = l_x, labels_y = l_y,
                    verbose = self.verbose )


        # Make sure that the output has the correct shape ------------------------------------
        if self.potentials:  # Return some dual potentials (= test functions) sampled on the input measures
            F, G = values
            return F.view_as(α), G.view_as(β)

        else:  # Return a scalar cost value
            if backend in ["online", "multiscale"]:  # KeOps backends return a single scalar value
                if B == 0: return values           # The user expects a scalar value
                else:      return values.view(-1)  # The user expects a "batch list" of distances

            else:  # "tensorized" backend returns a "batch vector" of values
                if B == 0: return values[0]  # The user expects a scalar value
                else:      return values     # The user expects a "batch vector" of distances
                

    def process_args(self, *args):
        if len(args) == 6:
            return args
        if len(args) == 4:
            α, x, β, y = args
            return None, α, x, None, β, y
        elif len(args) == 2:
            x, y = args
            α = self.generate_weights(x)
            β = self.generate_weights(y)
            return None, α, x, None, β, y
        else:
            raise ValueError("A SamplesLoss accepts two (x, y), four (α, x, β, y) or six (l_x, α, x, l_y, β, y)  arguments.")


    def generate_weights(self, x):
        if x.dim() == 2:  # 
            N = x.shape[0]
            return torch.ones(N).type_as(x) / N
        elif x.dim() == 3:
            B, N, _ = x.shape
            return torch.ones(B,N).type_as(x) / N
        else:
            raise ValueError("Input samples 'x' and 'y' should be encoded as (N,D) or (B,N,D) (batch) tensors.")


    def check_shapes(self, l_x, α, x, l_y, β, y):

        if α.dim() != β.dim(): raise ValueError("Input weights 'α' and 'β' should have the same number of dimensions.")
        if x.dim() != y.dim(): raise ValueError("Input samples 'x' and 'y' should have the same number of dimensions.")
        if x.shape[-1] != y.shape[-1]: raise ValueError("Input samples 'x' and 'y' should have the same last dimension.")

        if x.dim() == 2:  # No batch --------------------------------------------------------------------
            B = 0  # Batchsize
            N, D = x.shape  # Number of "i" samples, dimension of the feature space
            M, _ = y.shape  # Number of "j" samples, dimension of the feature space
            
            if α.dim() not in [1,2] :
                raise ValueError("Without batches, input weights 'α' and 'β' should be encoded as (N,) or (N,1) tensors.")
            elif α.dim() == 2:
                if α.shape[1] > 1: raise ValueError("Without batches, input weights 'α' should be encoded as (N,) or (N,1) tensors.")
                if β.shape[1] > 1: raise ValueError("Without batches, input weights 'β' should be encoded as (M,) or (M,1) tensors.")
                α, β = α.view(-1), β.view(-1)

            if l_x is not None:
                if l_x.dim() not in [1,2] :
                    raise ValueError("Without batches, the vector of labels 'l_x' should be encoded as an (N,) or (N,1) tensor.")
                elif l_x.dim() == 2:
                    if l_x.shape[1] > 1: raise ValueError("Without batches, the vector of labels 'l_x' should be encoded as (N,) or (N,1) tensors.")
                    l_x = l_x.view(-1)
                if len(l_x) != N : raise ValueError("The vector of labels 'l_x' should have the same length as the point cloud 'x'.")
            
            if l_y is not None:
                if l_y.dim() not in [1,2] :
                    raise ValueError("Without batches, the vector of labels 'l_y' should be encoded as an (M,) or (M,1) tensor.")
                elif l_y.dim() == 2:
                    if l_y.shape[1] > 1: raise ValueError("Without batches, the vector of labels 'l_y' should be encoded as (M,) or (M,1) tensors.")
                    l_y = l_y.view(-1)
                if len(l_y) != M : raise ValueError("The vector of labels 'l_y' should have the same length as the point cloud 'y'.")

            N2, M2 = α.shape[0], β.shape[0]

        elif x.dim() == 3:  # batch computation ---------------------------------------------------------
            B, N, D = x.shape  # Batchsize, number of "i" samples, dimension of the feature space
            B2,M, _ = y.shape  # Batchsize, number of "j" samples, dimension of the feature space
            if B != B2: raise ValueError("Samples 'x' and 'y' should have the same batchsize.")

            if α.dim() not in [2,3] :
                raise ValueError("With batches, input weights 'α' and 'β' should be encoded as (B,N) or (B,N,1) tensors.")
            elif α.dim() == 3:
                if α.shape[2] > 1: raise ValueError("With batches, input weights 'α' should be encoded as (B,N) or (B,N,1) tensors.")
                if β.shape[2] > 1: raise ValueError("With batches, input weights 'β' should be encoded as (B,M) or (B,M,1) tensors.")
                α, β = α.squeeze(-1), β.squeeze(-1)
            
            if l_x is not None: raise NotImplementedError('The "multiscale" backend has not been implemented with batches.')
            if l_y is not None: raise NotImplementedError('The "multiscale" backend has not been implemented with batches.')

            B2, N2 = α.shape
            B3, M2 = β.shape
            if B != B2: raise ValueError("Samples 'x' and weights 'α' should have the same batchsize.")
            if B != B3: raise ValueError("Samples 'y' and weights 'β' should have the same batchsize.")

        else:
            raise ValueError("Input samples 'x' and 'y' should be encoded as (N,D) or (B,N,D) (batch) tensors.")
        
        if N != N2: raise ValueError("Weights 'α' and samples 'x' should have compatible shapes.")
        if M != M2: raise ValueError("Weights 'β' and samples 'y' should have compatible shapes.")

        return B, N, M, D



"""Implements the (unbiased) Sinkhorn divergence between abstract measures.

.. math::
    \\text{S}_{\\varepsilon,\\rho}(\\alpha,\\beta) 
        ~&=~ \\text{OT}_{\\varepsilon,\\rho}(\\alpha, \\beta)
         ~-~\\tfrac{1}{2} \\text{OT}_{\\varepsilon,\\rho}(\\alpha, \\alpha)
         ~-~\\tfrac{1}{2} \\text{OT}_{\\varepsilon,\\rho}(\\beta, \\beta)
         ~+~ \\tfrac{\\varepsilon}{2} \| \\langle \\alpha, 1\\rangle - \\langle \\beta, 1\\rangle \|^2

where:

.. math::
    \\text{OT}_{\\varepsilon,\\rho}(\\alpha, \\beta)
    ~&=~ \\min_{\pi\geqslant 0} \\langle\, \pi\,,\, \\text{C} \,\\rangle
        ~+~\\varepsilon \, \\text{KL}(\pi,\\alpha\otimes\\beta) \\\\
        ~&+~\\rho \, \\text{KL}(\pi\,\mathbf{1},\\alpha)
        ~+~\\rho \, \\text{KL}(\pi^\intercal \,\mathbf{1},\\beta ) \\
    &=~ \\max_{b,a} -\\rho \\langle\, \\alpha \,,\, e^{-b/\\rho} - 1\,\\rangle
        -\\rho \\langle\, \\beta \,,\, e^{-a/\\rho} - 1\,\\rangle \\\\
        &-~
        \\epsilon \\langle\, \\alpha\otimes\\beta \,,\, e^{(b\oplus a - \\text{C})/\\epsilon} - 1\,\\rangle,

with a Kullback-Leibler divergence defined through:

.. math::
    \\text{KL}(\\alpha, \\beta)~=~
    \\langle \, \\alpha  \,,\, \\log \\tfrac{\\text{d}\\alpha}{\\text{d}\\beta} \,\\rangle
    ~-~ \\langle \, \\alpha  \,,\, 1 \,\\rangle
    ~+~ \\langle \, \\beta   \,,\, 1 \,\\rangle ~\geqslant~ 0.
"""

import numpy as np
import torch
from functools import partial

try:  # Import the keops library, www.kernel-operations.io
    from pykeops.torch import generic_logsumexp
    from pykeops.torch.cluster import grid_cluster, cluster_ranges_centroids, sort_clusters, from_matrix
    keops_available = True
except:
    keops_available = False
    
from .utils import scal, squared_distances, distances


# ==============================================================================
#                            ε-scaling heuristic
# ==============================================================================


def max_diameter(x, y):
    mins = torch.stack((x.min(dim=0)[0], y.min(dim=0)[0])).min(dim=0)[0]
    maxs = torch.stack((x.max(dim=0)[0], y.max(dim=0)[0])).max(dim=0)[0]
    diameter = (maxs-mins).norm().item()
    return diameter


def epsilon_schedule(p, diameter, blur, scaling):
    ε_s = [ diameter**p ] \
        + [ np.exp(e) for e in np.arange(p*np.log(diameter), p*np.log(blur), p*np.log(scaling)) ] \
        + [ blur**p ]
    return ε_s


def scaling_parameters( x, y, p, blur, reach, diameter, scaling):

    if diameter is None:
        D = x.shape[-1]
        diameter = max_diameter(x.view(-1,D), y.view(-1,D))

    ε   = blur**p
    ε_s = epsilon_schedule( p, diameter, blur, scaling )
    ρ   = None if reach is None else reach**p
    return diameter, ε, ε_s, ρ  


# ==============================================================================
#                              Sinkhorn loop
# ==============================================================================

def dampening(ε, ρ):
    return 1 if ρ is None else 1 / ( 1 + ε / ρ )


def log_weights(α):
    α_log = α.log()
    α_log[α <= 0] = -100000
    return α_log


class UnbalancedWeight(torch.nn.Module):
    def __init__(self, ε, ρ):
        super(UnbalancedWeight, self).__init__()
        self.ε, self.ρ = ε, ρ

    def forward(self, x):
        return (self.ρ + self.ε/2) * x

    def backward(self, g):
        return (self.ρ + self.ε)   * g


def sinkhorn_cost(ε, ρ, α, β, a_x, b_y, a_y, b_x, batch=False, debias=True, potentials=False):
    if potentials:  # Just return the dual potentials
        if debias:
            return b_x - a_x, a_y - b_y
        else:
            return b_x, a_y

    else:  # Actually compute the Sinkhorn divergence
        if debias:  # UNBIASED Sinkhorn divergence, S_ε(α,β) = OT_ε(α,β) - .5*OT_ε(α,α) - .5*OT_ε(β,β)
            if ρ is None:
                return scal( α, b_x - a_x, batch=batch ) + scal( β, a_y - b_y, batch=batch )
            else:
                return scal( α, UnbalancedWeight(ε, ρ)( (-a_x/ρ).exp() - (-b_x/ρ).exp() ), batch=batch ) \
                     + scal( β, UnbalancedWeight(ε, ρ)( (-b_y/ρ).exp() - (-a_y/ρ).exp() ), batch=batch )

        else:  # Classic, BIASED entropized Optimal Transport OT_ε(α,β)
            if ρ is None:
                return scal( α, b_x, batch=batch ) + scal( β, a_y, batch=batch )
            else:
                return scal( α, UnbalancedWeight(ε, ρ)( 1 - (-b_x/ρ).exp() ), batch=batch ) \
                     + scal( β, UnbalancedWeight(ε, ρ)( 1 - (-a_y/ρ).exp() ), batch=batch )


def sinkhorn_loop( softmin, α_logs, β_logs, C_xxs, C_yys, C_xys, C_yxs, ε_s, ρ, 
                   jumps=[], kernel_truncation=None, truncate=5, cost=None,
                   extrapolate=None, debias=True, last_extrapolation=True ):
    
    Nits = len(ε_s)
    if type(α_logs) is not list:
        α_logs, β_logs = [α_logs], [β_logs]
        if debias:
            C_xxs,  C_yys  = [C_xxs],  [C_yys]
        C_xys,  C_yxs  = [C_xys],  [C_yxs]


    torch.autograd.set_grad_enabled(False)

    k = 0  # Scale index; we start at the coarsest resolution available
    ε = ε_s[k] ; λ = dampening(ε, ρ)

    # Load the measures and cost matrices at the current scale:
    α_log, β_log = α_logs[k], β_logs[k]
    if debias:
        C_xx,  C_yy  = C_xxs[k],  C_yys[k]
    C_xy,  C_yx  = C_xys[k],  C_yxs[k]

    # Start with a decent initialization for the dual vectors:
    if debias:
        a_x = λ * softmin(ε, C_xx, α_log )  # OT(α,α)
        b_y = λ * softmin(ε, C_yy, β_log )  # OT(β,β)
    a_y = λ * softmin(ε, C_yx, α_log )  # OT(α,β) wrt. a
    b_x = λ * softmin(ε, C_xy, β_log )  # OT(α,β) wrt. b

    for i, ε in enumerate(ε_s):  # ε-scaling descent -----------------------

        λ = dampening(ε, ρ)  # ε has changed, so we should update λ too!

        # "Coordinate ascent" on the dual problems:
        if debias:
            at_x = λ * softmin(ε, C_xx, α_log + a_x/ε )  # OT(α,α)
            bt_y = λ * softmin(ε, C_yy, β_log + b_y/ε )  # OT(β,β)
        at_y = λ * softmin(ε, C_yx, α_log + b_x/ε )  # OT(α,β) wrt. a
        bt_x = λ * softmin(ε, C_xy, β_log + a_y/ε )  # OT(α,β) wrt. b

        # Symmetrized updates:
        if debias:
            a_x, b_y = .5 * ( a_x + at_x ), .5 * ( b_y + bt_y )  # OT(α,α), OT(β,β)
        a_y, b_x = .5 * ( a_y + at_y ), .5 * ( b_x + bt_x )  # OT(α,β) wrt. a, b


        if i in jumps:  # Jump from a coarse to a finer scale --------------

            if i == len(ε_s) - 1:  # Last iteration: just extrapolate!

                if debias:
                    C_xx_, C_yy_ = C_xxs[k+1], C_yys[k+1]
                C_xy_, C_yx_ = C_xys[k+1], C_yxs[k+1]

                last_extrapolation = False  # No need to re-extrapolate after the loop
                torch.autograd.set_grad_enabled( True )

            else:  # It's worth investing some time on kernel truncation...

                # Kernel truncation trick (described in Bernhard Schmitzer's 2016 paper),
                # that typically relies on KeOps' block-sparse routines:
                if debias:
                    C_xx_, _     = kernel_truncation( C_xx, C_xx, C_xxs[k+1], C_xxs[k+1],
                                                        a_x, a_x, ε, truncate=truncate,cost=cost)
                    C_yy_, _     = kernel_truncation( C_yy, C_yy, C_yys[k+1], C_yys[k+1],
                                                        b_y, b_y, ε, truncate=truncate,cost=cost)
                C_xy_, C_yx_ = kernel_truncation( C_xy, C_yx, C_xys[k+1], C_yxs[k+1],
                                                    b_x, a_y, ε, truncate=truncate,cost=cost)


            # Extrapolation for the symmetric problems:
            if debias:
                a_x = extrapolate( a_x, a_x, ε, λ, C_xx, α_log, C_xx_ )
                b_y = extrapolate( b_y, b_y, ε, λ, C_yy, β_log, C_yy_ )

            # The cross-updates should be done in parallel!
            a_y, b_x = extrapolate( a_y, b_x, ε, λ, C_yx, α_log, C_yx_ ), \
                       extrapolate( b_x, a_y, ε, λ, C_xy, β_log, C_xy_ )


            # Update the measure weights and cost "matrices":
            k = k+1
            α_log, β_log = α_logs[k], β_logs[k]
            if debias:
                C_xx,  C_yy  = C_xx_,  C_yy_
            C_xy,  C_yx  = C_xy_,  C_yx_

    torch.autograd.set_grad_enabled(True)

    if last_extrapolation:
        # Last extrapolation, to get the correct gradients:
        if debias:
            a_x = λ * softmin(ε, C_xx, (α_log + a_x/ε).detach() )
            b_y = λ * softmin(ε, C_yy, (β_log + b_y/ε).detach() )

        # The cross-updates should be done in parallel!
        a_y, b_x = λ * softmin(ε, C_yx, (α_log + b_x/ε).detach() ), \
                   λ * softmin(ε, C_xy, (β_log + a_y/ε).detach() )

    if debias:
        return a_x, b_y, a_y, b_x
    else:
        return None, None, a_y, b_x




"""Implements the (unbiased) Sinkhorn divergence between sampled measures."""

import numpy as np
import torch
from functools import partial

try:  # Import the keops library, www.kernel-operations.io
    from pykeops.torch import generic_logsumexp
    from pykeops.torch.cluster import grid_cluster, cluster_ranges_centroids
    from pykeops.torch.cluster import sort_clusters, from_matrix, swap_axes
    keops_available = True
except:
    keops_available = False
    
from .utils import scal, squared_distances, distances

from .sinkhorn_divergence import epsilon_schedule, scaling_parameters
from .sinkhorn_divergence import dampening, log_weights, sinkhorn_cost, sinkhorn_loop



# ==============================================================================
#                          backend == "tensorized"
# ==============================================================================

cost_routines = {
    1 : (lambda x,y : distances(x,y)),
    2 : (lambda x,y : squared_distances(x,y) / 2),
}

def softmin_tensorized(ε, C, f):
    B = C.shape[0]
    return - ε * ( f.view(B,1,-1) - C/ε ).logsumexp(2)

def sinkhorn_tensorized(α, x, β, y, p=2, blur=.05, reach=None, diameter=None, scaling=.5, cost=None, 
                        debias = True, potentials = False, **kwargs):
    
    B, N, D = x.shape
    _, M, _ = y.shape

    if cost is None:
        cost = cost_routines[p]
        
    C_xx, C_yy = ( cost( x, x.detach()), cost( y, y.detach()) ) if debias else (None, None)  # (B,N,N), (B,M,M)
    C_xy, C_yx = ( cost( x, y.detach()), cost( y, x.detach()) )  # (B,N,M), (B,M,N)


    diameter, ε, ε_s, ρ = scaling_parameters( x, y, p, blur, reach, diameter, scaling )

    a_x, b_y, a_y, b_x = sinkhorn_loop( softmin_tensorized, 
                                        log_weights(α), log_weights(β), 
                                        C_xx, C_yy, C_xy, C_yx, ε_s, ρ, debias=debias )

    return sinkhorn_cost(ε, ρ, α, β, a_x, b_y, a_y, b_x, batch=True, debias=debias, potentials=potentials)


# ==============================================================================
#                          backend == "online"
# ==============================================================================

cost_formulas = {
    1 : "Norm2(X-Y)",
    2 : "(SqDist(X,Y) / IntCst(2))",
}

def softmin_online(ε, C_xy, f_y, log_conv=None):
    x, y = C_xy
    # KeOps is pretty picky on the input shapes...
    return - ε * log_conv( x, y, f_y.view(-1,1), torch.Tensor([1/ε]).type_as(x) ).view(-1)


def keops_lse(cost, D):
    log_conv = generic_logsumexp("( B - (P * " + cost + " ) )",
                                 "A = Vi(1)",
                                 "X = Vi({})".format(D),
                                 "Y = Vj({})".format(D),
                                 "B = Vj(1)",
                                 "P = Pm(1)")
    return log_conv


def sinkhorn_online(α, x, β, y, p=2, blur=.05, reach=None, diameter=None, scaling=.5, cost=None, 
                    debias = True, potentials = False, **kwargs):
    
    N, D = x.shape
    M, _ = y.shape

    if cost is None: cost = cost_formulas[p]

    softmin = partial(softmin_online, log_conv=keops_lse(cost, D)) 

    # The "cost matrices" are implicitely encoded in the point clouds,
    # and re-computed on-the-fly:
    C_xx, C_yy = ( (x, x.detach()), (y, y.detach()) ) if debias else (None, None)
    C_xy, C_yx = ( (x, y.detach()), (y, x.detach()) )

    diameter, ε, ε_s, ρ = scaling_parameters( x, y, p, blur, reach, diameter, scaling )

    a_x, b_y, a_y, b_x = sinkhorn_loop( softmin,
                                        log_weights(α), log_weights(β), 
                                        C_xx, C_yy, C_xy, C_yx, ε_s, ρ, debias=debias )

    return sinkhorn_cost(ε, ρ, α, β, a_x, b_y, a_y, b_x, debias=debias, potentials=potentials)



# ==============================================================================
#                          backend == "multiscale"
# ==============================================================================


def softmin_multiscale(ε, C_xy, f_y, log_conv=None):
    x, y, ranges_x, ranges_y, ranges_xy = C_xy
    # KeOps is pretty picky on the input shapes...
    return - ε * log_conv( x, y, f_y.view(-1,1), torch.Tensor([1/ε]).type_as(x), ranges=ranges_xy ).view(-1)


def clusterize(α, x, scale=None, labels=None) :
    """
    Performs a simple 'voxelgrid' clustering on the input measure,
    putting points into cubic bins of size 'scale' = σ_c.
    The weights are summed, and the centroid position is that of the bin's center of mass.
    Most importantly, the "fine" lists of weights and points are *sorted*
    so that clusters are *contiguous in memory*: this allows us to perform
    kernel truncation efficiently on the GPU.

    If 
        [α_c, α], [x_c, x], [x_ranges] = clusterize(α, x, σ_c),
    then
        α_c[k], x_c[k] correspond to
        α[x_ranges[k,0]:x_ranges[k,1]], x[x_ranges[k,0]:x_ranges[k,1],:]
    """
    if labels is None and scale is None : # No clustering, single-scale Sinkhorn on the way...
        return [α], [x], []

    else : # As of today, only two-scale Sinkhorn is implemented:
        # Compute simple (voxel-like) class labels:
        x_lab = grid_cluster(x, scale) if labels is None else labels
        # Compute centroids and weights:
        ranges_x, x_c, α_c = cluster_ranges_centroids(x, x_lab, weights=α)
        # Make clusters contiguous in memory:
        (α, x), x_labels = sort_clusters( (α,x), x_lab)

        return [α_c, α], [x_c, x], [ranges_x]

def kernel_truncation( C_xy, C_yx, C_xy_, C_yx_, 
                       b_x, a_y, ε, truncate=None, cost=None, verbose=False):
    """Prunes out useless parts of the (block-sparse) cost matrices for finer scales.

    This is where our approximation takes place.
    To be mathematically rigorous, we should make several coarse-to-fine passes,
    making sure that we're not forgetting anyone. A good reference here is
    Bernhard Schmitzer's work: "Stabilized Sparse Scaling Algorithms for 
    Entropy Regularized Transport Problems, (2016)".
    """
    if truncate is None:
        return C_xy_, C_yx_
    else:
        x,  yd,   ranges_x,  ranges_y, _ = C_xy
        y,  xd,          _,         _, _ = C_yx
        x_, yd_, ranges_x_, ranges_y_, _ = C_xy_
        y_, xd_,         _,         _, _ = C_yx_

        with torch.no_grad():
            C      = cost(x, y)
            keep   = b_x.view(-1,1) + a_y.view(1,-1) > C - truncate*ε
            ranges_xy_ = from_matrix(ranges_x, ranges_y, keep)
            if verbose:
                ks, Cs = keep.sum(), C.shape[0]*C.shape[1]
                print("Keep {}/{} = {:2.1f}% of the coarse cost matrix.".format(
                    ks, Cs, 100*float(ks) / Cs ) )
    

        return (x_, yd_, ranges_x_, ranges_y_, ranges_xy_), \
               (y_, xd_, ranges_y_, ranges_x_, swap_axes(ranges_xy_))


def extrapolate_samples( b_x, a_y, ε, λ, C_xy, β_log, C_xy_, softmin=None ):
    yd = C_xy[1]   # Source points (coarse)
    x_ = C_xy_[0]  # Target points (fine)

    C = (x_, yd, None, None, None)  # "Rectangular" cost matrix, don't bother with ranges
    return λ * softmin(ε, C, (β_log + a_y/ε).detach() )


def sinkhorn_multiscale(α, x, β, y, p=2, blur=.05, reach=None, diameter=None, 
                        scaling=.5, truncate=5, cost=None, cluster_scale=None, 
                        debias = True, potentials = False,
                        labels_x = None, labels_y = None,
                        verbose=False, **kwargs):
    
    N, D = x.shape
    M, _ = y.shape

    if cost is None: cost = cost_formulas[p], cost_routines[p]
    cost_formula, cost_routine = cost[0], cost[1]

    softmin = partial(softmin_multiscale, log_conv=keops_lse(cost_formula, D)) 
    extrapolate = partial(extrapolate_samples, softmin=softmin)

    diameter, ε, ε_s, ρ = scaling_parameters( x, y, p, blur, reach, diameter, scaling )
    
    # Clusterize and sort our point clouds:
    if cluster_scale is None:
        cluster_scale = diameter / (np.sqrt(D) * 2000**(1/D))
    [α_c, α], [x_c, x], [ranges_x] = clusterize(α, x, scale=cluster_scale, labels=labels_x)
    [β_c, β], [y_c, y], [ranges_y] = clusterize(β, y, scale=cluster_scale, labels=labels_y)

    jumps = [ len(ε_s)-1 ]
    for i, ε in enumerate(ε_s[2:]):
        if cluster_scale**p > ε:
            jumps = [i+1]
            break
    

    if verbose: 
        print("{}x{} clusters, computed at scale = {:2.3f}".format(
              len(x_c), len(y_c), cluster_scale))
        print("Successive scales : ", ', '.join(["{:.3f}".format(x**(1/p)) for x in ε_s]))
        if jumps[0] >= len(ε_s)-1:
            print("Extrapolate from coarse to fine after the last iteration.")
        else:
            print("Jump from coarse to fine between indices {} (σ={:2.3f}) and {} (σ={:2.3f}).".format(
                jumps[0], ε_s[jumps[0]]**(1/p), jumps[0]+1, ε_s[jumps[0]+1]**(1/p)))


    # The input measures are stored at two levels: coarse and fine
    α_logs = [ log_weights(α_c), log_weights(α) ]
    β_logs = [ log_weights(β_c), log_weights(β) ]

    # We do the same [ coarse, fine ] decomposition for "cost matrices",
    # which are implicitely encoded as point clouds
    # + integer summation ranges, and re-computed on-the-fly:
    C_xxs = [ (x_c, x_c.detach(), ranges_x, ranges_x, None), 
              (  x,   x.detach(),     None,     None, None) ] if debias else None
    C_yys = [ (y_c, y_c.detach(), ranges_y, ranges_y, None), 
              (  y,   y.detach(),     None,     None, None) ] if debias else None
    C_xys = [ (x_c, y_c.detach(), ranges_x, ranges_y, None), 
              (  x,   y.detach(),     None,     None, None) ] 
    C_yxs = [ (y_c, x_c.detach(), ranges_y, ranges_x, None), 
              (  y,   x.detach(),     None,     None, None) ] 

    a_x, b_y, a_y, b_x = sinkhorn_loop( softmin,
                                        α_logs, β_logs, 
                                        C_xxs, C_yys, C_xys, C_yxs, ε_s, ρ,
                                        jumps=jumps,
                                        cost=cost_routine,
                                        kernel_truncation=partial(kernel_truncation, verbose=verbose),
                                        truncate=truncate,
                                        extrapolate=extrapolate, 
                                        debias = debias)

    return sinkhorn_cost(ε, ρ, α, β, a_x, b_y, a_y, b_x, debias=debias, potentials=potentials)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

import geomloss

setup(
    name='geomloss',
    version=geomloss.__version__,

    description='Geometric loss functions between point clouds, images and volumes.',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Jean Feydy',
    author_email='jean.feydy@ens.fr',

    python_requires='>=3',

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',

        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: MIT License',

        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',

        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='kernels optimal transport measure loss geometry',

    packages=['geomloss'],

    package_data={
        'geomloss': [
        ]
    },
    scripts=[],
    url='',
    license='LICENSE.txt',
    install_requires=[
        'numpy',
    ],

    extras_require={
        'full': ['pykeops[full]'],
    },
)



import torch

def scal(α, f, batch=False) :
    if batch:
        B = α.shape[0]
        return ( α.view(B,-1) * f.view(B,-1) ).sum(1)
    else:
        return torch.dot( α.view(-1), f.view(-1) )


class Sqrt0(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input):
        result = input.sqrt()
        result[input < 0] = 0
        ctx.save_for_backward(result)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        result, = ctx.saved_tensors
        grad_input = grad_output / (2*result)
        grad_input[result == 0] = 0
        return grad_input

def sqrt_0(x):
    return Sqrt0.apply(x)


def squared_distances(x, y):
    if x.dim() == 2:
        D_xx = (x*x).sum(-1).unsqueeze(1)  # (N,1)
        D_xy = torch.matmul( x, y.permute(1,0) )  # (N,D) @ (D,M) = (N,M)
        D_yy = (y*y).sum(-1).unsqueeze(0)  # (1,M)
    elif x.dim() == 3:  # Batch computation
        D_xx = (x*x).sum(-1).unsqueeze(2)  # (B,N,1)
        D_xy = torch.matmul( x, y.permute(0,2,1) )  # (B,N,D) @ (B,D,M) = (B,N,M)
        D_yy = (y*y).sum(-1).unsqueeze(1)  # (B,1,M)
    else:
        print("x.shape : ", x.shape)
        raise ValueError("Incorrect number of dimensions")

    return D_xx - 2*D_xy + D_yy


def distances(x, y):
    return sqrt_0( squared_distances(x,y) )



"""Implements kernel ("gaussian", "laplacian", "energy") norms between sampled measures.

.. math::
    \\text{Loss}(\\alpha,\\beta) 
        ~&=~ \\text{Loss}\\big( \sum_{i=1}^N \\alpha_i \,\delta_{x_i} \,,\, \sum_{j=1}^M \\beta_j \,\delta_{y_j} \\big) 
        ~=~ \\tfrac{1}{2} \|\\alpha-\\beta\|_k^2 \\\\
        &=~ \\tfrac{1}{2} \langle \\alpha-\\beta \,,\, k\star (\\alpha - \\beta) \\rangle \\\\
        &=~ \\tfrac{1}{2} \sum_{i=1}^N \sum_{j=1}^N  \\alpha_i \\alpha_j \cdot k(x_i,x_j) 
          + \\tfrac{1}{2} \sum_{i=1}^M \sum_{j=1}^M  \\beta_i \\beta_j \cdot k(y_i,y_j) \\\\
        &-~\sum_{i=1}^N \sum_{j=1}^M  \\alpha_i \\beta_j \cdot k(x_i,y_j)

where:

.. math::
    k(x,y)~=~\\begin{cases}
        \exp( -\|x-y\|^2/2\sigma^2) & \\text{if loss = ``gaussian''} \\\\
        \exp( -\|x-y\|/\sigma) & \\text{if loss = ``laplacian''} \\\\
        -\|x-y\| & \\text{if loss = ``energy''} \\\\
    \\end{cases}
"""

import numpy as np
import torch

try:  # Import the keops library, www.kernel-operations.io
    from pykeops.torch import generic_sum
    from pykeops.torch.cluster import grid_cluster, cluster_ranges_centroids, sort_clusters, from_matrix, swap_axes
    keops_available = True
except:
    keops_available = False

from .utils import scal, squared_distances, distances

class DoubleGrad(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        return input

    @staticmethod
    def backward(ctx, grad_output):
        return 2*grad_output

def double_grad(x):
    return DoubleGrad.apply(x)


# ==============================================================================
#                          backend == "tensorized"
# ==============================================================================

def gaussian_kernel(x, y, blur=.05):
    C2 = squared_distances(x / blur, y / blur)
    return (- .5 * C2 ).exp()

def laplacian_kernel(x, y, blur=.05):
    C = distances(x / blur, y / blur)
    return (- C ).exp()

def energy_kernel(x, y, blur=None):
    return - distances(x, y)

kernel_routines = {
    "gaussian" : gaussian_kernel,
    "laplacian": laplacian_kernel,
    "energy"   : energy_kernel,
}

def kernel_tensorized(α, x, β, y, blur=.05, kernel=None, name=None, potentials=False,**kwargs):
    
    B, N, D = x.shape
    _, M, _ = y.shape

    if kernel is None:
        kernel = kernel_routines[name]

    K_xx = kernel( double_grad(x), x.detach(), blur=blur)  # (B,N,N) tensor
    K_yy = kernel( double_grad(y), y.detach(), blur=blur)  # (B,M,M) tensor
    K_xy = kernel( x, y, blur=blur)                        # (B,N,M) tensor

    a_x = torch.matmul( K_xx, α.detach().unsqueeze(-1) ).squeeze(-1)  # (B,N,N) @ (B,N) = (B,N) 
    b_y = torch.matmul( K_yy, β.detach().unsqueeze(-1) ).squeeze(-1)  # (B,M,M) @ (B,M) = (B,M)
    b_x = torch.matmul( K_xy, β.unsqueeze(-1)          ).squeeze(-1)  # (B,N,M) @ (B,M) = (B,N) 

    
    if potentials:
        a_y = torch.matmul( K_xy.transpose(1,2), α.unsqueeze(-1)).squeeze(-1)  # (B,M,N) @ (B,N) = (B,M)
        return a_x - b_x, b_y - a_y

    else:  # Return the Kernel norm. N.B.: we assume that 'kernel' is symmetric:
        return .5 * (double_grad(α) * a_x).sum(1) \
             + .5 * (double_grad(β) * b_y).sum(1) \
             -  (α * b_x).sum(1)



# ==============================================================================
#                           backend == "online"
# ==============================================================================

kernel_formulas = {
    "gaussian" : ("Exp(-SqDist(X,Y) / IntCst(2))", True ),
    "laplacian": ("Exp(-Norm2(X-Y))",   True ),
    "energy"   : ("(-Norm2(X-Y))",      False),
}


def kernel_keops(kernel, α, x, β, y, potentials=False, ranges_xx = None, ranges_yy = None, ranges_xy = None):

    D = x.shape[1]
    kernel_conv = generic_sum( "(" + kernel + " * B)",   # Formula
                               "A = Vi(1)",              # Output:    a_i
                               "X = Vi({})".format(D),   # 1st input: x_i
                               "Y = Vj({})".format(D),   # 2nd input: y_j
                               "B = Vj(1)" )             # 3rd input: b_j
    
    a_x = kernel_conv(double_grad(x), x.detach(), α.detach().view(-1,1), ranges=ranges_xx)
    b_y = kernel_conv(double_grad(y), y.detach(), β.detach().view(-1,1), ranges=ranges_yy)
    b_x = kernel_conv(x, y, β.view(-1,1), ranges=ranges_xy)

    if potentials:
        a_y = kernel_conv(y, x, α.view(-1,1), ranges=swap_axes(ranges_xy))
        return a_x - b_x, b_y - a_y

    else:  # Return the Kernel norm. N.B.: we assume that 'kernel' is symmetric:
        return .5 * scal( double_grad(α), a_x ) \
             + .5 * scal( double_grad(β), b_y )  -  scal( α, b_x )
              


def kernel_preprocess(kernel, name, x, y, blur):
    if not keops_available:
        raise ImportError("The 'pykeops' library could not be loaded: " \
                        + "'online' and 'multiscale' backends are not available.")
    
    if kernel is None: kernel, rescale = kernel_formulas[name]
    else:              rescale = True
    
    # Center the point clouds just in case, to prevent numeric overflows:
    center = (x.mean(0, keepdim=True) + y.mean(0,  keepdim=True)) / 2
    x, y = x - center, y - center
    # Rescaling on x and y is cheaper than doing it for all pairs of points 
    if rescale : x, y = x / blur, y / blur
    
    return kernel, x, y


def kernel_online(α, x, β, y, blur=.05, kernel=None, name=None, potentials=False, **kwargs):

    kernel, x, y = kernel_preprocess(kernel, name, x, y, blur)
    return kernel_keops(kernel, α, x, β, y, potentials=potentials)


# ==============================================================================
#                          backend == "multiscale"
# ==============================================================================

def max_diameter(x, y):
    mins = torch.stack((x.min(dim=0)[0], y.min(dim=0)[0])).min(dim=0)[0]
    maxs = torch.stack((x.max(dim=0)[0], y.max(dim=0)[0])).max(dim=0)[0]
    diameter = (maxs-mins).norm().item()
    return diameter


def kernel_multiscale(α, x, β, y, blur=.05, kernel=None, name=None, 
                      truncate=5, diameter=None, cluster_scale=None, 
                      potentials=False, verbose=False, **kwargs):

    if truncate is None or name == "energy":
        return kernel_online( α, x, β, y, blur=blur, kernel=kernel, 
                              truncate=truncate, name=name, potentials=potentials, **kwargs )

    # Renormalize our point cloud so that blur = 1:
    kernel, x, y = kernel_preprocess(kernel, name, x, y, blur)

    # Don't forget to normalize the diameter too!
    if cluster_scale is None: 
        D = x.shape[-1]
        if diameter is None:
            diameter = max_diameter(x.view(-1,D), y.view(-1,D))
        else:
            diameter = diameter / blur
        cluster_scale = diameter / (np.sqrt(D) * 2000**(1/D))

    # Put our points in cubic clusters:
    cell_diameter = cluster_scale * np.sqrt( x.shape[1] )
    x_lab = grid_cluster(x, cluster_scale) 
    y_lab = grid_cluster(y, cluster_scale) 

    # Compute the ranges and centroids of each cluster:
    ranges_x, x_c, α_c = cluster_ranges_centroids(x, x_lab, weights=α)
    ranges_y, y_c, β_c = cluster_ranges_centroids(y, y_lab, weights=β)


    if verbose: 
        print("{}x{} clusters, computed at scale = {:2.3f}".format(
              len(x_c), len(y_c), cluster_scale))

    # Sort the clusters, making them contiguous in memory:
    (α, x), x_lab = sort_clusters( (α, x), x_lab)
    (β, y), y_lab = sort_clusters( (β, y), y_lab)

    with torch.no_grad():  # Compute our block-sparse reduction ranges:
        # Compute pairwise distances between clusters:
        C_xx = squared_distances( x_c, x_c)
        C_yy = squared_distances( y_c, y_c)
        C_xy = squared_distances( x_c, y_c)

        # Compute the boolean masks:
        keep_xx = ( C_xx <= (truncate + cell_diameter)**2 )
        keep_yy = ( C_yy <= (truncate + cell_diameter)**2 )
        keep_xy = ( C_xy <= (truncate + cell_diameter)**2 )

        # Compute the KeOps reduction ranges:
        ranges_xx = from_matrix(ranges_x, ranges_x, keep_xx)
        ranges_yy = from_matrix(ranges_y, ranges_y, keep_yy)
        ranges_xy = from_matrix(ranges_x, ranges_y, keep_xy)
    
    return kernel_keops(kernel, α, x, β, y, potentials=potentials,
                ranges_xx=ranges_xx, ranges_yy=ranges_yy, ranges_xy=ranges_xy)

